#include <unistd.h>
#include <stdlib.h>

#define __USE_XOPEN
#include <sys/poll.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <sys/mman.h>
#include <features.h>		/* for the glibc version number */
#if __GLIBC__ >= 2 && __GLIBC_MINOR >= 1
#include <netpacket/packet.h>
#include <net/ethernet.h>	/* the L2 protocols */
#else
#include <asm/types.h>
#include <linux/if_packet.h>
#include <linux/if_ether.h>	/* The L2 protocols */
#endif

#include <string.h>
#include <netinet/in.h>
#include <asm/system.h>

#include <packet.h>
#include <cleanup.h>
#include <capdev.h>
#include <capture.h>
#include <decode.h>
#include <preproc.h>
#include <plugin.h>

proc_decode_proto decode_proto;
proc_preproc_dispatch preproc_dispatch;
struct proto *ipproto=NULL;

/* Only one fd is used for all mmap() packet socket
 * I/O, we just add interface memberships as and when
 * we need them 
 */
int refcount=0;
int fd;
char *map;
size_t maplen;
struct tpacket_req req;
struct iovec *ring;
struct sockaddr_ll addr;

/* This is our own private data */
struct lnx_priv {
	int iface;
};

/* Initialise a capture process, we open the file and then
 * return our opaque private data structure to firestorm */
void *lnx_init(char *interface)
{
	struct lnx_priv *p=malloc(sizeof(struct lnx_priv));
	int i;

	if (!p) return NULL;

	if ( interface ) {
		printf("linux: interface specification not supported yet\n");
	}

	if ( !ipproto ) ipproto=decode_proto("ipv4");
	if ( !ipproto ) {
		free(p);
		return NULL;
	}

	if ( refcount )
		goto finalbit;

	/* Open the packet socket */
	if ( (fd=socket(PF_PACKET, SOCK_DGRAM, 0))<0 ) {
		printf("linux: socket(): %s", get_err());
		free(p);
		return NULL;
	}

	/* Setup socket for RX ring buffer */
	req.tp_block_size=4096;
	req.tp_frame_size=2048;
	req.tp_block_nr=64;
	req.tp_frame_nr=2*64;
	if ( setsockopt(fd,SOL_PACKET,PACKET_RX_RING,
		(char *)&req, sizeof(req)) ) {
		printf("linux: setsockopt(): %s", get_err());
		close(fd);
		free(p);
		return NULL;
	}

	/* mmap() the sucker */
	maplen=req.tp_block_size * req.tp_block_nr;
	if ( (map=mmap(NULL, maplen,
		PROT_READ|PROT_WRITE|PROT_EXEC,
		MAP_SHARED, fd, 0))==MAP_FAILED ) {
		printf("linux: mmap(): %s", get_err());
		close(fd);
		free(p);
		return NULL;
	}

	/* Allocate ring buffer iovecs */
	if ( !(ring=malloc(maplen)) ) {
		printf("linux: malloc(): %s", get_err());
		munmap(map, maplen);
		close(fd);
		free(p);
	}

	/* Setup our ring buffer */
	for(i=0; i<req.tp_frame_nr; i++) {
		ring[i].iov_base=(void *)((long)map)+(i*req.tp_frame_size);
		ring[i].iov_len=req.tp_frame_size;
	}

	/* bind the packet socket */
	memset((char *)&addr, 0, sizeof(addr));
	addr.sll_family=AF_PACKET;
	addr.sll_protocol=__constant_htons(0x03);
	addr.sll_ifindex=0;
	addr.sll_hatype=0;
	addr.sll_pkttype=0;
	addr.sll_halen=0;
	if ( bind(fd, (struct sockaddr *)&addr, sizeof(addr)) ) {
		printf("linux: bind(): %s", get_err());
		munmap(map, maplen);
		close(fd);
		return NULL;
	}

finalbit:
	return p;
}

void lnx_end(void *priv)
{
	struct lnx_priv *p=priv;
	if ( !p ) return;
	free(p);

	if ( !--refcount ) {
		munmap(map, maplen);
		close(fd);
	}
}

unsigned int lnx_go(void *priv, struct cap_s *c)
{
	struct tpacket_hdr *h;
	struct sockaddr_ll *sll;
	struct pollfd pfd;
	struct packet pkt;
	unsigned int packets=0;
	int i;

	memset(&pkt, 0, sizeof(pkt));
	pkt.capture=c;
	pkt.flags=FP_CLONE|FP_LIVE;

	for(i=0;;) {
		if ( c->state == CAP_STATE_STOP) return packets;

		/* If state gets set to stop during this while loop
		 * we will not see it untill the next poll... It's a
		 * corner case really.*/
		while(*(unsigned long*)ring[i].iov_base) {
			h=ring[i].iov_base;
			sll=(void *)h + TPACKET_ALIGN(sizeof(*h));

			/* Only IP */
			if (sll->sll_protocol!=__constant_htons(0x0800))
				goto nextpkt;

			/* build packet */
			pkt.time.tv_sec=h->tp_sec;
			pkt.time.tv_usec=h->tp_usec;
			pkt.len=h->tp_len;
			pkt.caplen=h->tp_snaplen;
			pkt.base=(void *)h + h->tp_mac;
			pkt.end=pkt.base+pkt.caplen;

			/* Decode and dispatch */
			pkt.layer[0].proto=ipproto;
			pkt.layer[0].h.raw=pkt.base;
			pkt.llen=1;
			pkt.rlen=0;
			pkt.present=0;
			if ( ipproto->decode(&pkt) ) {
				preproc_dispatch(&pkt);
			}

			packets++;
			
nextpkt:
			h->tp_status=0; mb();

			i=(i>=req.tp_frame_nr-1) ? 0 : i+1;
		}

		pfd.fd=fd;
		pfd.events=POLLIN|POLLERR;
		pfd.revents=0;
		poll(&pfd, 1, -1);
	}
	
	return packets;
}

int _PLUGIN_INIT (struct plugin_in *in, struct plugin_out *out)
{
	proc_capdev_add capdev_add;

	/* validate input */
	plugin_check(in, out);

	PLUGIN_ID("capture.linux", "Linux mmap() packet socket");
	PLUGIN_VERSION(1, 0);
	PLUGIN_AUTHOR("Gianni Tedesco", "gianni@ecsc.co.uk");
	PLUGIN_LICENSE("GPL");

	if ( !(capdev_add=in->import("capdev.add")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(decode_proto=in->import("decode.proto")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(preproc_dispatch=in->import("preproc.dispatch")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !capdev_add("linux", lnx_init, lnx_end, lnx_go) ) {
		return PLUGIN_ERR_FAIL;
	}
	
	return PLUGIN_ERR_OK;
}

int _PLUGIN_UNLOAD (int code) {
	return PLUGIN_ERR_OK;
}
