#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <netinet/in.h>
#include <errno.h>

#include <packet.h>
#include <matcher.h>
#include <template.h>
#include <plugin.h>

#define FBM_NORMAL	0
#define FBM_ALL		1
#define FBM_ANY		2
#define FBM_NOT		3

proc_template template_shortrange;

/* Mask off the bits we need in the frag_off field */
u_int16_t fbmask=__constant_htons(~IP_OFFMASK);

struct fragbits_priv {
	u_int8_t	mode;
	u_int16_t	fragbits;
};

/* id - Match the fragment ID */
int id_match(struct packet *p, void *priv, unsigned int l, int n)
{
	return 0;
}

proc_match_match id_validate(char *args, void **priv)
{
	if ( template_shortrange(args, priv) )
		return id_match;
	else
		return NULL;
}

/* sameip - Match IP packets whose saddr=daddr */
int sip_compare(void *p1, void *p2) { return MCMP_IDENTICAL; }
int sip_match(struct packet *p, void *priv, unsigned int l, int n)
{
	return n ^ ( p->layer[l].h.ip->saddr == p->layer[l].h.ip->daddr );
}
proc_match_match sip_validate(char *args, void **priv)
{
	*priv=NULL;
	return sip_match;
}


/* ttl - Match the time-to-live field of an IP packet */
int ttl_compare(void *p1, void *p2)  {return (p1==p2);}
int ttl_match(struct packet *p, void *priv, unsigned int l, int n)
{
	unsigned char ttl=(unsigned char)((unsigned int)priv&0xff);
	return n ^ ( p->layer[l].h.ip->ttl==ttl );
}
proc_match_match ttl_validate(char *args, void **priv)
{
	unsigned long int val;

	if ( (val=strtoul(args, NULL, 10))==ULONG_MAX &&
		errno==ERANGE ) return NULL;

	/* Its a char */
	if ( val>255 ) return NULL;

	/* Store it directly in the pointer, hehe */
	*((unsigned int *)priv)=(unsigned int)val&0xff;
	
	return ttl_match;
}

/* fragbits - Match the fragmentation flag bits (in frag_off) */
int fb_match(struct packet *p, void *priv, unsigned int l, int n)
{
	struct fragbits_priv *x=priv;
	u_int16_t ipfb=p->layer[l].h.ip->frag_off&fbmask;

	switch ( x->mode ) {
		case FBM_NORMAL:
			return n ^ (ipfb == x->fragbits);
		case FBM_NOT:
			return n ^ !(ipfb & x->fragbits);
		case FBM_ALL:
			return n ^ ((ipfb & x->fragbits) == x->fragbits);
		case FBM_ANY:
			return n ^ (ipfb & x->fragbits);
		default:
			return 0;
	}
}

int fb_compare(void *p1, void *p2)
{
	struct fragbits_priv *f1=p1;
	struct fragbits_priv *f2=p2;
	
	if ( f1->mode==f2->mode && f1->fragbits==f2->fragbits )
		return MCMP_IDENTICAL;
	else
		return MCMP_DIFFER;
}

proc_match_match fb_validate(char *args, void **priv)
{
	struct fragbits_priv *fb;
	char *tmp;
	
	if ( !(fb=calloc(1, sizeof(*fb))) ) {
		return NULL;
	}

	for(tmp=args; *tmp; tmp++) {
		switch(*tmp) {
			case 'd':
			case 'D':
				fb->fragbits |= IP_DF;
				break;
			case 'm':
			case 'M':
				fb->fragbits |= IP_MF;
				break;
			case 'r':
			case 'R':
			case 'c':
			case 'C':
				fb->fragbits |= IP_CE;
				break;
			case '!':
				fb->mode=FBM_NOT;
				break;
			case '+':
				fb->mode=FBM_ANY;
				break;
			case '*':
				fb->mode=FBM_ALL;
				break;
			case ' ':
			case '\t':
			case ',':
				break;
			default:
				free(fb);
				return 0;
		}
	}
	
	fb->fragbits=htons(fb->fragbits);
	*priv=(void *)fb;
	return fb_match;
}


int _PLUGIN_INIT (struct plugin_in *in, struct plugin_out *out)
{
	int ok=0;
	proc_matcher_add matcher_add;
	
	plugin_check(in, out);
	
	PLUGIN_ID("match.ip", "IP matching routines");
	PLUGIN_VERSION(2, 0);
	PLUGIN_AUTHOR("Gianni Tedesco", "gianni@ecsc.co.uk");
	PLUGIN_LICENSE("GPL");

	if ( !(matcher_add=in->import("matcher.add")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(template_shortrange=in->import("template.shortrange")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	ok+=matcher_add("fragbits",
		fb_validate,
		fb_compare,
		in->import("template.ptr"));
	ok+=matcher_add("sameip",
		sip_validate,
		sip_compare,
		in->import("template.ptr"));
	ok+=matcher_add("ttl",
		ttl_validate,
		ttl_compare,
		NULL);
	ok+=matcher_add("id",
		id_validate,
		NULL,
		in->import("template.ptr"));
	
	return (ok) ? PLUGIN_ERR_OK : PLUGIN_ERR_FAIL;
}

int _PLUGIN_UNLOAD (int code) {
	return PLUGIN_ERR_OK;
}
